@@ -11,6 +11,27 @@
 *   01/04/05   DGP   Added entsymlookup().
 *   02/09/05   DGP   Unified error reporting using logerror.
 *   03/16/05   DGP   Change operation of symparse.
+*   04/26/05   DGP   Use pscanbuf in place of inbuf.
+*   06/09/05   DGP   Added opencopy/closecopy for INSERT.
+*   06/10/05   DGP   Added RMT support.
+*   12/15/05   RMS   MINGW changes.
+*   10/31/06   DGP   Change "/" expression to return zero. ie: PXD ,/
+*   07/30/07   DGP   Change mktemp to mkstemp.
+*   05/19/09   DGP   If RMT line is a duplicate, don't add to list.
+*   10/16/09   DGP   Pad Hollerith literal with blanks.
+*   04/30/10   DGP   Added sequence checking option.
+*   06/17/10   DGP   Changed tokscan to allow tokens beginning with '='.
+*   11/01/10   DGP   Added CTSS Common mode.
+*   12/15/10   DGP   Added line number to temporary file.
+*                    Added file sequence for errors.
+*   12/16/10   DGP   Added a unified source read function readsource().
+*   12/16/10   DGP   Added a unified source write function writesource().
+*   12/17/10   DGP   Added boolean/relocatable error.
+*   03/25/11   DGP   Allow (we ignore) "$" on INSERT filenames.
+*   03/30/11   DGP   Allow multiple include directories.
+*   11/11/11   DGP   Allow INSERT files to have "inc" or "fap" extension.
+*   11/16/11   DGP   In tokscan return qualifer with token.
+*   06/19/12   DGP   Fix scantok bug.
 *	
 ***********************************************************************/
 
@@ -19,6 +40,12 @@
 #include <ctype.h>
 #include <string.h>
 #include <time.h>
+#include <fcntl.h>
+#ifndef WIN32
+#include <unistd.h>
+#else
+#include <io.h>
+#endif
 
 #include "asmdef.h"
 #include "asmdmem.h"
@@ -30,7 +57,10 @@ extern int absolute;		/* In absolute section */
 extern int radix;		/* Number scanner radix */
 extern int linenum;		/* Source line number */
 extern int errcount;		/* Number of errors in assembly */
+extern int warncount;		/* Number of warnings in assembly */
 extern int errnum;		/* Number of pass 2 errors for current line */
+extern int warnnum;		/* Index into warnline array */
+extern int lastsequ;		/* Last sequence number */
 extern int exprtype;		/* Expression type */
 extern int p1errcnt;		/* Number of pass 0/1 errors */
 extern int p1erralloc;		/* Number of pass 0/1 errors allocated */
@@ -53,20 +83,392 @@ extern int rboolexpr;		/* RBOOL expression */
 extern int lboolexpr;		/* LBOOL expression */
 extern int fapmode;		/* FAP assembly mode */
 extern int headcount;		/* Number of entries in headtable */
+extern int filenum;		/* INSERT file index */
+extern int fileseq;		/* INSERT file sequence */
+extern int ctsscommon;		/* CTSS Common */
+extern int incldircnt;		/* Include directory count */
 
 extern char inbuf[MAXLINE];	/* The input buffer for the scanners */
 extern char qualtable[MAXQUALS][MAXSYMLEN+2]; /* The QUALifier table */
 extern char errline[10][120];	/* Pass 2 error lines for current statment */
+extern char warnline[10][120];	/* Pass 2 warning lines for current statment */
 extern char headtable[MAXHEADS];/* HEAD table */
+extern char *incldir[MAXINCLUDES]; /* -I include directories */
 
 extern SymNode *addextsym;	/* Added symbol for externals */
 extern SymNode *symbols[MAXSYMBOLS];/* The Symbol table */
 extern SymNode *entsymbols[MAXSYMBOLS];/* The Entry Symbol table */
 extern SymNode *freesymbols;	/* Reusable symbols nodes */
 extern XrefNode *freexrefs;	/* Reusable xref nodes */
+extern RMTSeq *rmthead;		/* Pointer to start of remote sequences */
+extern RMTSeq *rmttail;		/* Pointer to end of remote sequences */
 extern ErrTable p1error[MAXERRORS];/* The pass 0/1 error table */
+extern INSERTfile files[MAXASMFILES];/* INSERT file descriptors */
+extern char *pscanbuf;		/* Pointer for tokenizers */
 
 static char errtmp[120];	/* Error print string */
+#ifdef DEBUGSYM
+static int ssdebug;
+#endif
+
+/***********************************************************************
+* readsource - Read source mode, line number and text from the temp file.
+***********************************************************************/
+
+int
+readsource (FILE *fd, int *inmode, int *linenum, int *filenum, char *inbuf)
+{
+#ifdef WIN32
+   int i;
+   char temp[32];
+#endif
+#if defined(DEBUGP1RDR) || defined(DEBUGP2RDR)
+   int flag = FALSE;
+#endif
+
+#ifdef WIN32
+   /*
+   ** Stupid WinBLOWS can't handle inter mixed binary/ascii reads....
+   ** Even though we fdopen in binary mode.
+   */
+
+   for (i = 0; i < 8; i++)
+   {
+      if ((temp[i] = fgetc(fd)) == EOF)
+         return (-1);
+   }
+   temp[8] = 0;
+   sscanf (temp, "%x", inmode);
+   for (i = 0; i < 8; i++)
+   {
+      if ((temp[i] = fgetc(fd)) == EOF)
+         return (-1);
+   }
+   temp[8] = 0;
+   sscanf (temp, "%d", linenum);
+#else
+   if (fread(inmode, 1, 4, fd) != 4)
+      return (-1);
+   if (fread(linenum, 1, 4, fd) != 4)
+      return (-1);
+#endif
+
+   if (fgets(inbuf, MAXLINE, fd) == NULL)
+      return (-1);
+
+   *filenum = (*inmode & NESTMASK) >> NESTSHFT;
+
+#ifdef DEBUGP1RDR
+   if (inpass == 010) flag = TRUE;
+#endif
+#ifdef DEBUGP2RDR
+   if (inpass == 020) flag = TRUE;
+#endif
+#if defined(DEBUGP1RDR) || defined(DEBUGP2RDR)
+   if (flag)
+   {
+      fprintf (stderr, "linenum = %d, inmode = %06x, filenum = %d\n",
+	      *linenum, *inmode, *filenum);
+      fprintf (stderr, "inbuf = %s", inbuf);
+   }
+#endif
+   return (0);
+}
+
+/***********************************************************************
+* writesource - Write source to the temp file or seq.
+***********************************************************************/
+
+void
+writesource (FILE *fd, int rmtflg, int mode, int num, char *buf, char *loc)
+{
+   if (rmtflg)
+   {
+      rmtseqadd (mode, buf);
+   }
+   else
+   {
+#ifdef WIN32
+      char temp[16];
+#endif
+
+      mode |= ((filenum ? fileseq : 0) << NESTSHFT) & NESTMASK;
+
+#ifdef WIN32
+      /*
+      ** Stupid WinBLOWS can't handle inter mixed binary/ascii reads....
+      ** Even though we fdopen in binary mode.
+      */
+
+      sprintf (temp, "%08x", mode);
+      fwrite (temp, 1, 8, fd);
+      sprintf (temp, "%08d", num);
+      fwrite (temp, 1, 8, fd);
+#else
+      fwrite (&mode, 1, 4, fd);
+      fwrite (&num, 1, 4, fd);
+#endif
+
+      fputs (buf, fd);
+#ifdef DEBUGP0OUT
+      fprintf (stderr, "%s(%06X) %d:%s", loc, mode, num, buf);
+#endif
+   }
+}
+
+#ifdef WIN32
+/***********************************************************************
+* mkstemp - Brain dead support for WinBlows.
+***********************************************************************/
+
+int mkstemp (char *template)
+{
+   int fd;
+   char tmpnam[1024];
+
+   strcpy (tmpnam, mktemp(template));
+   fd = open (tmpnam, O_CREAT | O_TRUNC | O_RDWR, 0666);
+   strcpy (template, tmpnam);
+   return (fd);
+}
+#endif
+
+/***********************************************************************
+* opencopy - Open a copy INSERT
+***********************************************************************/
+
+FILE *
+opencopy (char *bp, FILE *tmpfd, char *etcbuf)
+{
+   FILE *infd;
+   char *cp;
+   static int tmpndx = 3;
+   int tmpdes;
+   int curfile;
+   int i, found;
+   int fapext;
+   char tname[1024];
+
+   while (isspace (*bp)) bp++;
+   cp = bp;
+   if (*cp == '$') cp++;
+   while (!isspace (*bp))
+   {
+      if (isupper(*bp)) *bp = tolower(*bp);
+      bp++;
+   }
+   *bp = '\0';
+
+   found = FALSE;
+   fapext = FALSE;
+   i = 0;
+   while (!found)
+   {
+      if (incldircnt && incldir[i])
+      {
+	 sprintf (tname, "%s/%s.%s", incldir[i], cp, fapext ? "fap" : "inc");
+      }
+      else
+      {
+	 sprintf (tname, "%s.%s", cp, fapext ? "fap" : "inc");
+      }
+#ifdef DEBUGP0RDR
+      fprintf (stderr, "opencopy: linenum = %d\n", linenum);
+      fprintf (stderr, "   insert = %s\n", tname);
+      fprintf (stderr, "   etcbuf = %s", etcbuf);
+#endif
+
+      if ((infd = fopen (tname, "r")) == NULL)
+      {
+	 if (!fapext)
+	 {
+	    fapext = TRUE;
+	 }
+	 else
+	 {
+	    fapext = FALSE;
+	    i++;
+	    if (i == incldircnt)
+	    {
+	       char tmp[1024];
+
+	       sprintf (tmp, "asm7090: Can't open input file for INSERT: %s",
+			tname);
+	       perror (tmp);
+	       exit (1);
+	    }
+	 }
+      }
+      else
+         found = TRUE;
+   }
+
+   curfile = filenum;
+   files[curfile].fd = tmpfd;
+   files[curfile].linenum = linenum;
+   files[curfile].lastseq = lastsequ;
+#if defined(WIN32)
+   sprintf (files[curfile].tmpname, "a%d%s", tmpndx++, TEMPSPEC);
+#else
+   sprintf (files[curfile].tmpname, "/tmp/a%d%s", tmpndx++, TEMPSPEC);
+#endif
+   strcpy (files[curfile].etcline, etcbuf);
+   etcbuf[0] = '\0';
+   filenum++;
+   fileseq++;
+
+   if (filenum >= MAXASMFILES)
+   {
+      fprintf (stderr, "asm7090: Too many INSERT files\n");
+      exit (1);
+   }
+   linenum = 0;
+   lastsequ = 0;
+   if ((tmpdes = mkstemp (files[curfile].tmpname)) < 0)
+   {
+      perror ("asm7090: Can't mkstemp for INSERT");
+      exit (1);
+   }
+
+   if ((tmpfd = fdopen (tmpdes, "w+")) == NULL)
+   {
+      perror ("asm7090: Can't open temporary file for INSERT");
+      exit (1);
+   }
+
+   detab (infd, tmpfd, "insert");
+   fclose (infd);
+   if (fseek (tmpfd, 0, SEEK_SET) < 0)
+   {
+      perror ("asm7090: Can't rewind temp file");
+      exit (1);
+   }
+   return (tmpfd);
+}
+
+/***********************************************************************
+* closecopy - Close a copy INSERT
+***********************************************************************/
+
+FILE *
+closecopy (FILE *tmpfd, char *etcbuf)
+{
+   if (filenum)
+   {
+      filenum--;
+      fclose (tmpfd);
+      unlink (files[filenum].tmpname);
+      linenum = files[filenum].linenum;
+      lastsequ = files[filenum].lastseq;
+      strcpy (etcbuf, files[filenum].etcline);
+#ifdef DEBUGP0RDR
+      fprintf (stderr, "closecopy: linenum = %d\n", linenum);
+      fprintf (stderr, "   etcbuf = %s", etcbuf);
+#endif
+      return (files[filenum].fd);
+   }
+   return (NULL);
+}
+
+/***********************************************************************
+* rmtseqadd - Add a line to the remote sequence pool.
+***********************************************************************/
+
+void
+rmtseqadd (int mode, char *line)
+{
+   RMTSeq *new;
+
+#if 0
+   new = rmthead;
+   while (new)
+   {
+      if (!strcmp(new->line, line))
+          return;
+       new = new->next;
+   }
+#endif
+
+   if ((new = (RMTSeq *)DBG_MEMGET (sizeof (RMTSeq))) == NULL)
+   {
+      fprintf (stderr, "asm7090: Unable to allocate memory\n");
+      exit (ABORT);
+   }
+
+   memset (new, '\0', sizeof (RMTSeq));
+
+   new->linemode = mode | RMTSEQ;
+   strcpy (new->line, line);
+
+   if (!rmthead) rmthead = new;
+   if (rmttail) rmttail->next = new;
+   rmttail = new;
+
+}
+
+/***********************************************************************
+* rmtseqget - Get remote sequence pool line for processing.
+***********************************************************************/
+
+char *
+rmtseqget (int *mode, char *line)
+{
+   if (rmthead)
+   {
+      RMTSeq *lclptr;
+
+      rmthead->linemode |= RMTSEQ;
+      lclptr = rmthead;
+      rmthead = rmthead->next;
+      strcpy (line, lclptr->line);
+      *mode = lclptr->linemode;
+      DBG_MEMREL (lclptr);
+      return (line);
+   }
+
+   rmttail = NULL;
+   return (NULL);
+}
+
+/***********************************************************************
+* chkerror - Check to see if the Pass 1 error has been recorded.
+***********************************************************************/
+
+static int
+chkerror (int line)
+{
+   int i;
+
+   for (i = 0; i < p1errcnt; i++)
+      if (p1error[i].errorline == line)
+         return (TRUE);
+   return (FALSE);
+}
+
+/***********************************************************************
+* logwarning - Log a warning for printing.
+***********************************************************************/
+
+void
+logwarning (char *warnmsg)
+{
+#ifdef DEBUGERROR
+   fprintf (stderr, "logwarning: linenum = %d, warncount = %d, inpass = %o\n",
+	   linenum, warncount, inpass);
+   fprintf (stderr, "   p1errcnt = %d, p1erralloc = %d, warnnum = %d\n",
+	   p1errcnt, p1erralloc, warnnum);
+   fprintf (stderr, "   warnmsg = %s\n", warnmsg);
+#endif
+
+   if (!chkerror (linenum))
+   {
+      warncount++;
+      if (inpass == 020)
+      {
+	 strcpy (warnline[warnnum++], warnmsg);
+      }
+   }
+}
 
 /***********************************************************************
 * logerror - Log an error for printing.
@@ -76,12 +478,13 @@ void
 logerror (char *errmsg)
 {
 #ifdef DEBUGERROR
-   printf ("logerror: linenum = %d, errcount = %d, inpass = %o\n",
+   fprintf (stderr, "logerror: linenum = %d, errcount = %d, inpass = %o\n",
 	   linenum, errcount, inpass);
-   printf ("   p1errcnt = %d, p1erralloc = %d, errnum = %d\n",
+   fprintf (stderr, "   p1errcnt = %d, p1erralloc = %d, errnum = %d\n",
 	   p1errcnt, p1erralloc, errnum);
-   printf ("   errmsg = %s\n", errmsg);
+   fprintf (stderr, "   errmsg = %s\n", errmsg);
 #endif
+
    if (!chkerror (linenum))
    {
       errcount++;
@@ -97,6 +500,7 @@ logerror (char *errmsg)
 	    p1erralloc++;
 	 }
 	 p1error[p1errcnt].errorline = linenum;
+	 p1error[p1errcnt].errorfile = (filenum ? fileseq : 0);
 	 strcpy (p1error[p1errcnt].errortext, errmsg);
 	 p1errcnt++;
       }
@@ -113,7 +517,7 @@ Parse_Error (int cause, int state, int defer)
    char errorstring[256];
 
 #ifdef DEBUGPARSEERROR
-   printf (
+   fprintf (stderr, 
    "Parse_Error: pass = %o, defer = %s, cause = %d, state = %d, linenum = %d\n",
 	 inpass, defer ? "TRUE " : "FALSE", cause, state, linenum);
 #endif
@@ -149,6 +553,11 @@ Parse_Error (int cause, int state, int defer)
    {
       strcpy (errorstring, "Divide by zero");
    }
+   else if (cause == EXPR_ERROR)
+   {
+      if (inpass != 020) return;
+      strcpy (errorstring, "Illegal use of boolean");
+   }
 
    sprintf (errtmp, "%s in expression", errorstring);
    logerror (errtmp);
@@ -160,7 +569,8 @@ Parse_Error (int cause, int state, int defer)
 ***********************************************************************/
 
 char *
-tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
+tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term,
+         int usequaltable)
 {
    int index;
    toktyp tt;
@@ -170,7 +580,7 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
    char c;
 
 #ifdef DEBUGTOK
-   printf ("tokscan: Entered: buf = %s", buf);
+   fprintf (stderr, "tokscan: Entered: buf = %s", buf);
 #endif
 
    /*
@@ -179,11 +589,16 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
 
    while (*buf && isspace(*buf))
    {
+#ifdef DEBUGTOK
+      fprintf (stderr, "   *buf = %c(%X)\n", *buf, *buf);
+      fprintf (stderr, "   buf = %p, pscanbuf = %p\n", buf, pscanbuf);
+#endif
+
       /*
       ** We stop scanning on a newline or when we're passed the operands.
       */
 
-      if (*buf == '\n' || (buf - inbuf >= NOOPERAND))
+      if (*buf == '\n' || (buf - pscanbuf >= NOOPERAND))
       {
 	 t[0] = '\0';
 	 *token = t;
@@ -195,19 +610,19 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
    }
 
    /*
-   ** Check for "strange" opcodes, convert to PZE
+   ** Check for "strange" opcodes.
    */
 
-   if ((buf-inbuf) >= OPCSTART && (buf-inbuf) < VARSTART)
+   if ((buf-pscanbuf) >= OPCSTART && (buf-pscanbuf) < VARSTART)
    {
 #ifdef DEBUGTOKNULLOP
-      printf ("tokscan: buf = %s", buf);
+      fprintf (stderr, "tokscan: buf = %s", buf);
 #endif
-      if (   !strncmp (buf, "...", 3)
-          || !strncmp (buf, "***", 3))
+      if (!strncmp (buf, "...", 3) ||
+          !strncmp (buf, "***", 3))
       {
 #ifdef DEBUGTOKNULLOP
-	 printf ("  null opcode ... found\n");
+	 fprintf (stderr, "  null opcode ... found\n");
 #endif
 	 strcpy (t, "PZE");
 	 *tokentype = SYM;
@@ -217,16 +632,28 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
 	 *term = *(buf+index);
 	 return (buf+index);
       }
+      else if (*buf == '=')
+      {
+	 char *bp = buf;
+	 while (*buf && !isspace(*buf)) buf++;
+	 index = buf - bp;
+	 strncpy (t, bp, index);
+	 t[index] = '\0';
+	 *tokentype = SYM;
+	 *tokenvalue = 0;
+	 *token = t;
+	 *term = *(bp+index);
+	 return (bp+index);
+      }
    }
 
-#if 0
    /*
    ** If nil operand, return type
    */
 
-   if (!strncmp (buf, "**", 2))
+   if (!strncmp (buf, "--", 2))
    {
-      strcpy (t, "**");
+      strcpy (t, "--");
       index = 2;
       if (radix == 8)
 	 *tokentype = OCTNUM;
@@ -237,7 +664,6 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
       *term = *(buf+index);
       return (buf+index);
    }
-#endif
 
    /*
    ** Call the scanner to get the token
@@ -245,13 +671,14 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
 
    index = 0;
 #ifdef DEBUGTOK
-   printf ("  calling scanner: buf = %s\n", buf);
+   fprintf (stderr, "  calling scanner: buf = %s\n", buf);
 #endif
-   tt = Scanner (buf, &index, &val, t, q, &c, inpass == 020 ? FALSE : TRUE); 
+   tt = Scanner (buf, &index, &val, t, q, &c, inpass == 020 ? FALSE : TRUE,
+                 usequaltable); 
 #ifdef DEBUGTOK
-   printf ("  token = %s, tokentype = %d, term = %02x, val = %d\n",
+   fprintf (stderr, "  token = %s, tokentype = %d, term = %02x, val = %d\n",
 	    t, tt, c, val);
-   printf (" index = %d, *(buf+index) = %02X\n",  index, *(buf+index));
+   fprintf (stderr, " index = %d, *(buf+index) = %02X\n",  index, *(buf+index));
 #endif
 
    /*
@@ -261,7 +688,23 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
    if (tt == DECNUM || tt == OCTNUM)
       c = *(buf+index);
    if (c == ',') index++;
-   *token = t;
+   if (q[0])
+   {
+      if (q[0] == '0')
+      {
+         strcpy (q, "$");
+      }
+      else
+      {
+         strcat (q, "$");
+      }
+      strcat (q, t);
+      *token = q;
+   }
+   else
+   {
+      *token = t;
+   }
    *tokentype = tt;
    *tokenvalue = val;
    *term = c;
@@ -275,8 +718,12 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
 void
 freexref (XrefNode *xr)
 {
+#ifdef DEBUGMALLOCS
+   DBG_MEMREL (xr);
+#else
    xr->next = freexrefs;
    freexrefs = xr;
+#endif
 }
 
 /***********************************************************************
@@ -286,8 +733,12 @@ freexref (XrefNode *xr)
 void
 freesym (SymNode *sr)
 {
+#ifdef DEBUGMALLOCS
+   DBG_MEMREL (sr);
+#else
    sr->next = freesymbols;
    freesymbols = sr;
+#endif
 }
 
 /***********************************************************************
@@ -341,10 +792,21 @@ symlookup (char *sym, char *qual, int add, int xref)
    char lksym[80];
 
 #ifdef DEBUGSYM
-   printf ("symlookup: Entered: sym = '%s', qual = '%s'\n", sym, qual);
-   printf ("   add = %s, xref = %s\n",
+#ifdef DEBUGSYMVAL
+   ssdebug = !strcmp (sym, "LPFLAG");
+#else
+   ssdebug = TRUE;
+#endif
+   if (ssdebug)
+   {
+      fprintf (stderr, "symlookup: Entered: sym = '%s', qual = '%s'\n",
+	    sym, qual);
+      fprintf (stderr,
+	    "   add = %s, xref = %s, inpass = %03o, pc = %05o, line = %d\n",
 	    add ? "TRUE" : "FALSE",
-	    xref ? "TRUE" : "FALSE");
+	    xref ? "TRUE" : "FALSE",
+	    inpass, pc, linenum);
+   }
 #endif
 
    /*
@@ -356,7 +818,8 @@ symlookup (char *sym, char *qual, int add, int xref)
       if (!add) return (NULL);
 
 #ifdef DEBUGSYM
-      printf ("   add symbol at top\n");
+      if (ssdebug)
+	 fprintf (stderr, "   add symbol at top\n");
 #endif
       if (freesymbols)
       {
@@ -374,12 +837,14 @@ symlookup (char *sym, char *qual, int add, int xref)
       symbols[0]->value = pc;
       symbols[0]->line = linenum;
 #ifdef DEBUGXREF
-      printf ("symlookup: sym = %s defline = %d\n", sym, linenum);
+      fprintf (stderr, "symlookup: sym = %s defline = %d\n", sym, linenum);
 #endif
       if (absolute || absmod)
-	 symbols[0]->relocatable = FALSE;
+	 symbols[0]->flags &= ~RELOCATABLE;
       else
-	 symbols[0]->relocatable = TRUE;
+	 symbols[0]->flags |= RELOCATABLE;
+      if (inpass == 000)
+	 symbols[0]->flags |= P0SYM;
       symbolcount++;
       return (symbols[0]);
    }
@@ -397,7 +862,7 @@ symlookup (char *sym, char *qual, int add, int xref)
    {
       mid = (up - lo) / 2 + lo;
 #ifdef DEBUGSYMSEARCH
-      printf ("   mid = %d, last = %d\n", mid, last);
+      fprintf (stderr, "   mid = %d, last = %d\n", mid, last);
 #endif
       if (symbolcount == 1)
 	 done = TRUE;
@@ -405,7 +870,7 @@ symlookup (char *sym, char *qual, int add, int xref)
 
       sprintf (tblsym, "%s$%s", symbols[mid]->symbol, symbols[mid]->qualifier);
 #ifdef DEBUGSYMSEARCH
-      printf ("   tblsym = '%s', lksym = '%s'\n", tblsym, lksym);
+      fprintf (stderr, "   tblsym = '%s', lksym = '%s'\n", tblsym, lksym);
 #endif
       r = strcmp (tblsym, lksym);
 
@@ -418,7 +883,28 @@ symlookup (char *sym, char *qual, int add, int xref)
 	 SymNode *symp = symbols[mid];
 
 	 if (add)
-	    return (NULL); /* must be a duplicate */
+	 {
+	    if (!(symbols[mid]->flags & P0SYM))
+	       return (NULL); /* must be a duplicate */
+
+#ifdef DEBUGSYM
+	    if (ssdebug)
+	       fprintf (stderr, "   update symbol\n");
+#endif
+	    symbols[mid]->flags &= ~P0SYM;
+	    symbols[mid]->value = pc;
+	    symbols[mid]->line = linenum;
+	    if (absolute || absmod)
+	       symbols[mid]->flags &= ~RELOCATABLE;
+	    else
+	       symbols[mid]->flags |= RELOCATABLE;
+	 }
+
+#ifdef DEBUGSYM
+	 if (ssdebug)
+	    fprintf (stderr, "   found symbol, value = %05o\n",
+		  symbols[mid]->value);
+#endif
 
 	 /*
 	 ** If xref mode, chain a line reference
@@ -439,7 +925,8 @@ symlookup (char *sym, char *qual, int add, int xref)
 	       exit (ABORT);
 	    }
 #ifdef DEBUGXREF
-	    printf ("symlookup: sym = %s refline = %d\n", sym, linenum);
+	    fprintf (stderr, "symlookup: sym = %s refline = %d\n",
+		     sym, linenum);
 #endif
 	    memset (new, '\0', sizeof(XrefNode));
 	    new->next = NULL;
@@ -481,7 +968,8 @@ symlookup (char *sym, char *qual, int add, int xref)
       SymNode *new;
 
 #ifdef DEBUGSYM
-      printf ("   add new symbol\n");
+      if (ssdebug)
+	 fprintf (stderr, "   add new symbol\n");
 #endif
       /*
       ** Add to table, allocate storage and initialize symbol
@@ -508,14 +996,16 @@ symlookup (char *sym, char *qual, int add, int xref)
       strcpy (new->symbol, sym);
       strcpy (new->qualifier, qual);
 #ifdef DEBUGXREF
-      printf ("symlookup: sym = %s defline = %d\n", sym, linenum);
+      fprintf (stderr, "symlookup: sym = %s defline = %d\n", sym, linenum);
 #endif
       new->value = pc;
       new->line = linenum;
       if (absolute || absmod)
-	 new->relocatable = FALSE;
+	 new->flags &= ~RELOCATABLE;
       else
-	 new->relocatable = TRUE;
+	 new->flags |= RELOCATABLE;
+      if (inpass == 000)
+	 new->flags |= P0SYM;
 
       /*
       ** Insert pointer in sort order.
@@ -561,10 +1051,20 @@ entsymlookup (char *sym, char *qual, int add, int xref)
    char lksym[80];
 
 #ifdef DEBUGSYM
-   printf ("entsymlookup: Entered: sym = '%s', qual = '%s'\n", sym, qual);
-   printf ("   add = %s, xref = %s\n",
-	    add ? "TRUE" : "FALSE",
-	    xref ? "TRUE" : "FALSE");
+#ifdef DEBUGSYMVAL
+   ssdebug = !strcmp (sym, "LPFLAG");
+#else
+   ssdebug = TRUE;
+#endif
+   if (ssdebug)
+   {
+      fprintf (stderr, "entsymlookup: Entered: sym = '%s', qual = '%s'\n",
+	       sym, qual);
+      fprintf (stderr, "   add = %s, xref = %s, inpass = %03o, linenum = %d\n",
+	       add ? "TRUE" : "FALSE",
+	       xref ? "TRUE" : "FALSE",
+	       inpass, linenum);
+   }
 #endif
 
    /*
@@ -576,7 +1076,8 @@ entsymlookup (char *sym, char *qual, int add, int xref)
       if (!add) return (NULL);
 
 #ifdef DEBUGSYM
-      printf ("add symbol at top\n");
+      if (ssdebug)
+	 fprintf (stderr, "add symbol at top\n");
 #endif
       if (freesymbols)
       {
@@ -595,12 +1096,12 @@ entsymlookup (char *sym, char *qual, int add, int xref)
       entsymbols[0]->value = pc;
       entsymbols[0]->line = linenum;
 #ifdef DEBUGXREF
-      printf ("entsymlookup: sym = %s defline = %d\n", sym, linenum);
+      fprintf (stderr, "entsymlookup: sym = %s defline = %d\n", sym, linenum);
 #endif
       if (absolute || absmod)
-	 entsymbols[0]->relocatable = FALSE;
+	 entsymbols[0]->flags &= ~RELOCATABLE;
       else
-	 entsymbols[0]->relocatable = TRUE;
+	 entsymbols[0]->flags |= RELOCATABLE;
       entsymbolcount++;
       return (entsymbols[0]);
    }
@@ -618,7 +1119,7 @@ entsymlookup (char *sym, char *qual, int add, int xref)
    {
       mid = (up - lo) / 2 + lo;
 #ifdef DEBUGSYMSEARCH
-      printf (" mid = %d, last = %d\n", mid, last);
+      fprintf (stderr, " mid = %d, last = %d\n", mid, last);
 #endif
       if (entsymbolcount == 1)
 	 done = TRUE;
@@ -627,7 +1128,7 @@ entsymlookup (char *sym, char *qual, int add, int xref)
       sprintf (tblsym, "%s$%s",
 	       entsymbols[mid]->symbol, entsymbols[mid]->qualifier);
 #ifdef DEBUGSYMSEARCH
-      printf ("   tblsym = '%s', lksym = '%s'\n", tblsym, lksym);
+      fprintf (stderr, "   tblsym = '%s', lksym = '%s'\n", tblsym, lksym);
 #endif
       r = strcmp (tblsym, lksym);
 
@@ -661,7 +1162,8 @@ entsymlookup (char *sym, char *qual, int add, int xref)
 	       exit (ABORT);
 	    }
 #ifdef DEBUGXREF
-	    printf ("entsymlookup: sym = %s refline = %d\n", sym, linenum);
+	    fprintf (stderr, "entsymlookup: sym = %s refline = %d\n",
+		  sym, linenum);
 #endif
 	    memset (new, '\0', sizeof(XrefNode));
 	    new->next = NULL;
@@ -703,7 +1205,8 @@ entsymlookup (char *sym, char *qual, int add, int xref)
       SymNode *new;
 
 #ifdef DEBUGSYM
-      printf ("add new symbol\n");
+      if (ssdebug)
+	 fprintf (stderr, "add new symbol\n");
 #endif
       /*
       ** Add to table, allocate storage and initialize symbol
@@ -730,14 +1233,14 @@ entsymlookup (char *sym, char *qual, int add, int xref)
       strcpy (new->symbol, sym);
       strcpy (new->qualifier, qual);
 #ifdef DEBUGXREF
-      printf ("entsymlookup: sym = %s defline = %d\n", sym, linenum);
+      fprintf (stderr, "entsymlookup: sym = %s defline = %d\n", sym, linenum);
 #endif
       new->value = pc;
       new->line = linenum;
       if (absolute || absmod)
-	 new->relocatable = FALSE;
+	 new->flags &= ~RELOCATABLE;
       else
-	 new->relocatable = TRUE;
+	 new->flags |= RELOCATABLE;
 
       /*
       ** Insert pointer in sort order.
@@ -779,7 +1282,7 @@ symlocate (char *sym, int *count)
    int r;
 
 #ifdef DEBUGSYMLOC
-   printf ("symlocate: Entered: sym = '%s'\n", sym);
+   fprintf (stderr, "symlocate: Entered: sym = '%s'\n", sym);
 #endif
 
    /*
@@ -803,7 +1306,8 @@ symlocate (char *sym, int *count)
       if (r == 0)
       {
 #ifdef DEBUGSYMLOC
-	 printf ("   mid = %d, qual = %s\n", mid, symbols[mid]->qualifier);
+	 fprintf (stderr, "   mid = %d, qual = %s\n",
+		  mid, symbols[mid]->qualifier);
 #endif
 	 /*
 	 ** Check if other symbols with different qualifiers above and below.
@@ -814,7 +1318,8 @@ symlocate (char *sym, int *count)
 	        strcmp (symbols[mid]->qualifier, symbols[lo-1]->qualifier))
 	 {
 #ifdef DEBUGSYMLOC
-	    printf ("   lo = %d, qual = %s\n", lo-1, symbols[lo-1]->qualifier);
+	    fprintf (stderr, "   lo = %d, qual = %s\n",
+		     lo-1, symbols[lo-1]->qualifier);
 #endif
 	    lo--;
 	 }
@@ -823,7 +1328,8 @@ symlocate (char *sym, int *count)
 	        strcmp (symbols[mid]->qualifier, symbols[up-1]->qualifier))
 	 {
 #ifdef DEBUGSYMLOC
-	    printf ("   up = %d, qual = %s\n", up-1, symbols[up-1]->qualifier);
+	    fprintf (stderr, "   up = %d, qual = %s\n",
+		     up-1, symbols[up-1]->qualifier);
 #endif
 	    up++;
 	 }
@@ -834,7 +1340,8 @@ symlocate (char *sym, int *count)
 
 	 *count = up - lo + 1;
 #ifdef DEBUGSYMLOC
-	 printf ("   lo = %d, up = %d, count = %d\n", lo, up, *count);
+	 fprintf (stderr, "   lo = %d, up = %d, count = %d\n",
+		  lo, up, *count);
 #endif
          return (symbols[mid]);
       }
@@ -852,7 +1359,7 @@ symlocate (char *sym, int *count)
 ***********************************************************************/
 
 tokval 
-symparse (char *sym, char *qual, int *relo, int defer, int pcinc)
+symparse (char *sym, char *qual, int *relo, int *boolsym, int defer, int pcinc)
 {
    SymNode *s;
    tokval value;
@@ -864,13 +1371,15 @@ symparse (char *sym, char *qual, int *relo, int defer, int pcinc)
 
    s = NULL;
    addextsym = NULL;
+   *boolsym = FALSE;
    relocatable = *relo;
 
    xref = (((inpass & 070) == 010) && defer == FALSE) ? FALSE : TRUE;
 
 #ifdef DEBUGSYMPARSE
-   printf ("symparse: Entered: sym = '%s', qual = '%s'\n", sym, qual);
-   printf ("   defer = %s, xref = %s\n",
+   fprintf (stderr, "symparse: Entered: sym = '%s', qual = '%s'\n", sym, qual);
+   fprintf (stderr, "   exprtype = %03o, defer = %s, xref = %s\n",
+	    exprtype,
 	    defer ? "TRUE" : "FALSE",
 	    xref ? "TRUE" : "FALSE");
 #endif
@@ -982,10 +1491,10 @@ symparse (char *sym, char *qual, int *relo, int defer, int pcinc)
 	    {
 	       s->value = 0;
 	       if (absmod || absolute)
-		  s->relocatable = FALSE;
+		  s->flags &= ~RELOCATABLE;
 	       else
-		  s->relocatable = TRUE;
-	       s->external = TRUE;
+		  s->flags |= RELOCATABLE;
+	       s->flags |= EXTERNAL;
 	       addextsym = s;
 	       goto SYM_FOUND;
 	    }
@@ -1009,6 +1518,9 @@ symparse (char *sym, char *qual, int *relo, int defer, int pcinc)
    {
 
    SYM_FOUND:
+#ifdef DEBUGSYMPARSE
+      fprintf (stderr, "    SYM_FOUND:\n");
+#endif
       /*
       ** If in Boolean expression handle appropriately
       */
@@ -1020,7 +1532,7 @@ symparse (char *sym, char *qual, int *relo, int defer, int pcinc)
 	 ** Only BOOL data allowed in a boolean expression
 	 */
 
-	 if (!s->bool)
+	 if (!(s->flags & BOOL))
 	 {
 	    sprintf (errtmp, "Invalid symbol for BOOL expression: %s", sym);
 	    logerror (errtmp);
@@ -1035,17 +1547,19 @@ symparse (char *sym, char *qual, int *relo, int defer, int pcinc)
 	 {
 #endif
 #ifdef DEBUGBOOLSYM
-            printf (
+            fprintf (stderr, 
 	       "symparse: BOOL sym = %s, val = %o, lrbool = %s, rbool = %s\n",
 		  sym, s->value,
-		  s->lrbool ? "TRUE" : "FALSE",
-		  s->rbool ? "TRUE" : "FALSE");
+		  s->flags & LRBOOL ? "TRUE" : "FALSE",
+		  s->flags & RBOOL ? "TRUE" : "FALSE");
 #endif
 	    value = s->value;
-	    if (s->lrbool)
+	    if (s->flags & BOOLBITS)
+	       *boolsym = TRUE;
+	    if (s->flags & LRBOOL)
 	    {
-	       if (s->rbool) rboolexpr = TRUE;
-	       else          lboolexpr = TRUE;
+	       if (s->flags & RBOOL) rboolexpr = TRUE;
+	       else                  lboolexpr = TRUE;
 	    }
 #if 0
 	 }
@@ -1058,12 +1572,29 @@ symparse (char *sym, char *qual, int *relo, int defer, int pcinc)
 
       else
       {
+#if 0
+	 if (s->flags & BOOL)
+	 {
+	    sprintf (errtmp, "Invalid symbol for BOOL expression: %s", sym);
+	    logerror (errtmp);
+	    value = 0;
+	 }
+#endif
+	 if (s->flags & BOOLBITS)
+	    *boolsym = TRUE;
+
 	 /* Mark REF as being used */
-	 if (s->external && s->value == 0)
-	    s->relocatable = !absolute;
-	 if (s->relocatable) relocatable++;
+	 if ((s->flags & EXTERNAL) && s->value == 0)
+	 {
+	    if (absolute)
+	       s->flags &= ~RELOCATABLE;
+	    else
+	       s->flags |= RELOCATABLE;
+	 }
+	 if (s->flags & RELOCATABLE) relocatable++;
+	 else if (ctsscommon && (s->flags & COMMON)) relocatable++;
 	 value = s->value;
-	 if (s->external)
+	 if (s->flags & EXTERNAL)
 	 {
 	    /* Don't relocate since it's extern tail */
 	    if (value == 0) relocatable = 100;
@@ -1092,7 +1623,7 @@ exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
    tokval lval = 0;
 
 #ifdef DEBUGEXPR
-   printf ("exprscan: entered: bp = %s\n", bp);
+   fprintf (stderr, "exprscan: entered: bp = %s\n", bp);
 #endif
    cp = bp;
 
@@ -1110,7 +1641,7 @@ exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
    ** If no operands, return EOL
    */
 
-   if (spaces && (bp - inbuf >= NOOPERAND))
+   if (spaces && (bp - pscanbuf >= NOOPERAND))
    {
       bp = cp;
       *tterm = '\n';
@@ -1120,43 +1651,60 @@ exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
    }
 
 #ifdef DEBUGEXPR
-   printf ("   bp = %s\n", bp);
+   fprintf (stderr, "   bp = %s\n", bp);
 #endif
 
-#if 0
    /*
    ** If nil operand, return type
    */
 
-   if (!strncmp (bp, "**", 2))
+   if (!strncmp (bp, "--", 2))
    {
       index = 2;
       *tterm = *(bp+index);
       *val = lval;
       *relo = lrelo;
 #ifdef DEBUGEXPR
-      printf ("   null op: term = %c\n", *tterm);
+      fprintf (stderr, "   null op: term = %c\n", *tterm);
 #endif
       if (*tterm == ',') index++;
       return (bp+index);
    }
-#endif
 
    /*
-   ** Else, if literal then scan it off and put in symbol table
+   ** If literal then scan it off and put in symbol table
    */
 
-   else if (*bp == LITERALSYM)
+   if (*bp == LITERALSYM)
    {
       SymNode *s;
       char *cp;
 
-      cp = bp;
-      while (*cp && !isspace(*cp)) cp++;
+      cp = bp+1;
+      if (*cp == 'H')
+      {
+         while (*cp && *cp >= ' ' && ((cp-bp) < 8)) cp++;
+	 if (((cp-bp) < 8) || *cp < ' ')
+	    while ((cp-bp) < 8)
+	    {
+	       *cp++ = ' ';
+	    }
+      }
+      else if (*cp == 'V')
+      {
+	 while (*cp && !isspace(*cp)) cp++;
+      }
+      else
+      {
+	 while (*cp && *cp != ',' && !isspace(*cp)) cp++;
+      }
+      *tterm = *cp;
+      index = cp - bp;
       *cp = '\0';
 
 #ifdef DEBUGLIT
-      printf ("exprscan: pc = %o, litpool = %d, litorigin = %o, inpass = %o\n",
+      fprintf (stderr,
+	       "exprscan: pc = %o, litpool = %d, litorigin = %o, inpass = %o\n",
 	       pc, litpool, litorigin, inpass);
 #endif
       if (!(s = symlookup (bp, qualtable[qualindex], FALSE, TRUE)))
@@ -1180,8 +1728,9 @@ exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
 	       s->value = pgmlength++;
 	 }
       }
+      *cp = *tterm;
       lval = s->value;
-      if (s->relocatable) lrelo = TRUE;
+      if (s->flags & RELOCATABLE) lrelo = TRUE;
    }
 
    /*
@@ -1214,7 +1763,7 @@ exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
       }
       else if (*bp == '/' && (*(bp+1) == ',' || isspace (*(bp+1))))
       {
-	 lval = (exprtype & BOOLVALUE) ? 0777777 : 077777 ;
+	 lval = -1;
 	 index = 1;
 	 *tterm = *(bp+index);
 	 *val = lval;
@@ -1225,9 +1774,9 @@ exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
       lval = Parser (bp, &lrelo, &index, entsize, defer, pcinc);
    }
 #ifdef DEBUGEXPR
-   printf (" index = %d, *(bp+index) = %02X\n",
+   fprintf (stderr, " index = %d, *(bp+index) = %02X\n",
 	 index, *(bp+index));
-   printf (" lrelo = %d, lval = %d\n", lrelo, lval);
+   fprintf (stderr, " lrelo = %d, lval = %d\n", lrelo, lval);
 #endif
    
    /*
@@ -1256,7 +1805,7 @@ condexpr (char *bp, int *val, char *term)
    int ret;
 
 #ifdef DEBUGCONDEXPR
-   printf ("condexpr: linenum = %d: bp = %s", linenum, bp);
+   fprintf (stderr, "condexpr: linenum = %d: bp = %s", linenum, bp);
 #endif
 
    /*
@@ -1322,7 +1871,7 @@ condexpr (char *bp, int *val, char *term)
          rel = EQ;
       }
 #ifdef DEBUGCONDEXPR
-      printf ("   lbp = %s, rel = %d, rbp = %s\n", lbp, rel, rbp);
+      fprintf (stderr, "   lbp = %s, rel = %d, rbp = %s\n", lbp, rel, rbp);
 #endif
 
       /* 
@@ -1363,7 +1912,7 @@ condexpr (char *bp, int *val, char *term)
 	    *cp = '\0';
 	 }
 #ifdef DEBUGCONDEXPR
-	 printf ("   lch = '%s', rch = '%s'\n", lch, rch);
+	 fprintf (stderr, "   lch = '%s', rch = '%s'\n", lch, rch);
 #endif
 	 /*
 	 ** Compare string tokens
@@ -1396,7 +1945,7 @@ condexpr (char *bp, int *val, char *term)
 	 else
 	    rval = 0;
 #ifdef DEBUGCONDEXPR
-	 printf ("   lval = %d, rval = %d\n", lval, rval);
+	 fprintf (stderr, "   lval = %d, rval = %d\n", lval, rval);
 #endif
 
 	 /*
@@ -1409,7 +1958,7 @@ condexpr (char *bp, int *val, char *term)
       }
 
       /*
-      ** Set return relation based on requesed compare
+      ** Set return relation based on requested compare
       */
 
       if (rel == EQ && r == 0) ret = TRUE;
@@ -1417,8 +1966,8 @@ condexpr (char *bp, int *val, char *term)
       else if (rel == LT && r < 0) ret = TRUE;
       else ret = FALSE;
 #ifdef DEBUGCONDEXPR
-      printf ("   val = %d, ret = %s\n", r, ret ? "TRUE" : "FALSE");
-      printf ("   term = %c(%x), bp = %s\n", *term, *term, bp);
+      fprintf (stderr, "   val = %d, ret = %s\n", r, ret ? "TRUE" : "FALSE");
+      fprintf (stderr, "   term = %c(%x), bp = %s\n", *term, *term, bp);
 #endif
       *val = ret;
    }
