@@ -1,22 +1,21 @@
 /***********************************************************************
 * Scan tables
-*     This module defines the tables used by the
-*     scanner. Two tables are defined: CHARTABLE -
-*     the character translation table used to find
-*     the translate value of the input character* and
-*     SCNTBL - the finite state machine upon which the
-*     scanner operates. SCNTBL is an array of pointers
-*     (indexed by state number) to state transition
-*     lists. Each entry of the state list contains a
-*     field for the comparand character, the next
-*     state and a scan action which applies when the
-*     comparand character matches the translated input
-*     character. Each state transition list ends with
-*     a default entry which is used when the current
-*     input character does not match any characters in
-*     the list. Also in this module are equates
-*     (character translate equates) and macros (scan
-*     state element) to generate the two scan tables.
+*     This module defines the tables used by the scanner. Two tables are
+*     defined:
+*
+*     chartable - the character translation table used to find the
+*                 translate value of the input character
+*     scntbl - the finite state machine upon which the scanner operates.
+*
+*     scntbl is an array of pointers (indexed by state number) to state
+*     transition lists. Each entry of the state list contains a field for
+*     the comparand character, the next state and a scan action which
+*     applies when the comparand character matches the translated input
+*     character. Each state transition list ends with a default entry
+*     which is used when the current input character does not match any
+*     characters in the list. Also in this module are equates (character
+*     translate equates) and macros (scan state element) to generate the
+*     two scan tables.
 *
 ***********************************************************************/
 /*
@@ -54,7 +53,7 @@ static char chartable[] = {
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >0C->0F */
 
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >10->13 */
-        GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >14->17 */
+        GRAPHIC,EOLN   ,GRAPHIC,GRAPHIC,  /*            >14->17 */
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >18->1B */
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >1C->1F */
 
@@ -81,7 +80,7 @@ static char chartable[] = {
         MINUS,  SLASH,  GRAPHIC,GRAPHIC,  /* - /        >60->63 */
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >64->67 */
         GRAPHIC,GRAPHIC,GRAPHIC,COMMA,    /*       ,    >68->6B */
-        SPECIAL,SPECIAL,RELOP,  SPECIAL,  /* % _ > ?    >6C->6F */
+        SPECIAL,ALPHA,  RELOP,  SPECIAL,  /* % _ > ?    >6C->6F */
 
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >70->73 */
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >74->77 */
@@ -130,7 +129,7 @@ static char chartable[] = {
 #else
         EOLN,   EOFS,   GRAPHIC,GRAPHIC,  /*            >00->03 */
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >04->07 */
-        GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >08->0B */
+        GRAPHIC,GRAPHIC,EOLN   ,GRAPHIC,  /*            >08->0B */
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >0C->0F */
 
         GRAPHIC,GRAPHIC,GRAPHIC,GRAPHIC,  /*            >10->13 */
@@ -156,7 +155,7 @@ static char chartable[] = {
         ALPHA,  ALPHA,  ALPHA,  ALPHA,    /* P Q R S    >50->53 */
         ALPHA,  ALPHA,  ALPHA,  ALPHA,    /* T U V W    >54->57 */
         ALPHA,  ALPHA,  ALPHA,  OPERAT,   /* X Y Z [    >58->5B */
-        SPECIAL,SPECIAL,SPECIAL,SPECIAL,  /* \ ] ^ _    >5C->5F */
+        SPECIAL,OPERAT, SPECIAL,ALPHA,    /* \ ] ^ _    >5C->5F */
 
         SPECIAL,ALPHA,  BB,     ALPHA,    /* ` a b c    >60->63 */
         ALPHA,  EE,     ALPHA,  ALPHA,    /* d e f g    >64->67 */
@@ -197,6 +196,7 @@ static short int S1[] = {
         ENTRY(EE,2,EAT_ACTION|MOVE_ACTION),
         ENTRY(PERIOD,2,EAT_ACTION|MOVE_ACTION),
         ENTRY(LPARN,2,EAT_ACTION|MOVE_ACTION),
+        ENTRY(RPARN,2,EAT_ACTION|MOVE_ACTION),
         ENTRY(DOLLAR,3,EAT_ACTION|MOVE_ACTION),
         ENTRY(ODIGIT,4,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
         ENTRY(DIGIT,4,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
@@ -279,6 +279,7 @@ static short int S7[] = {
         ENTRY(EE,8,EAT_ACTION|MOVE_ACTION),
         ENTRY(PERIOD,8,EAT_ACTION|MOVE_ACTION),
         ENTRY(LPARN,8,EAT_ACTION|MOVE_ACTION),
+        ENTRY(RPARN,8,EAT_ACTION|MOVE_ACTION),
         ENTRY(DIGIT,8,EAT_ACTION|MOVE_ACTION),
         ENTRY(ODIGIT,9,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
         ENTRY(EOLN,10,BUILD_ACTION),
@@ -327,6 +328,7 @@ static short int S10[] = {
         ENTRY(BB,11,EAT_ACTION|MOVE_ACTION),
         ENTRY(EE,11,EAT_ACTION|MOVE_ACTION),
         ENTRY(LPARN,11,EAT_ACTION|MOVE_ACTION),
+        ENTRY(RPARN,11,EAT_ACTION|MOVE_ACTION),
         ENTRY(DIGIT,13,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
         ENTRY(ODIGIT,13,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
         ENTRY(PERIOD,14,EAT_ACTION|MOVE_ACTION),
@@ -422,6 +424,7 @@ static short int S16[] = {
 static short int S17[] = {
         ENTRY(DIGIT,17,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
         ENTRY(ODIGIT,17,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(BB,28,EAT_ACTION|MOVE_ACTION),
         ENTRY(DEFAULT,18,BUILD_ACTION)
 };
 
@@ -439,9 +442,10 @@ static short int S19[] = {
 	ENTRY(PLUS,20,EAT_ACTION),
 	ENTRY(MINUS,20,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
 	ENTRY(BB,19,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(EE,32,EAT_ACTION|MOVE_ACTION),
 	ENTRY(DIGIT,21,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
 	ENTRY(ODIGIT,21,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
-        ENTRY(DEFAULT,0,ERROR_ACTION)
+        ENTRY(DEFAULT,22,BUILD_ACTION)
 };
 
 /*
@@ -459,6 +463,7 @@ static short int S20[] = {
 static short int S21[] = {
         ENTRY(DIGIT,21,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
         ENTRY(ODIGIT,21,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(EE,32,EAT_ACTION|MOVE_ACTION),
         ENTRY(DEFAULT,22,BUILD_ACTION)
 };
 
@@ -473,10 +478,13 @@ static short int S22[] = {
 ** State 23 - recognize null symbol "**"
 */
 static short int S23[] = {
+	ENTRY(EOLN,0,BUILD_ACTION),
 	ENTRY(SPACE,0,BUILD_ACTION),
 	ENTRY(COMMA,0,BUILD_ACTION),
 	ENTRY(GRAPHIC,0,BUILD_ACTION),
 	ENTRY(PLUS,0,BUILD_ACTION),
+	ENTRY(LPARN,0,BUILD_ACTION),
+	ENTRY(OPERAT,0,BUILD_ACTION),
         ENTRY(DEFAULT,24,BACK_ACTION|BUILD_ACTION)
 };
 
@@ -513,9 +521,86 @@ static short int S27[] = {
         ENTRY(DEFAULT,0,NULL_ACTION)
 };
 
+/*
+** State 28 - recognize nnEnnBnn format
+*/
+static short int S28[] = {
+	ENTRY(PLUS,29,EAT_ACTION|BUILD_ACTION),
+	ENTRY(MINUS,29,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(BB,28,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(DIGIT,30,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(ODIGIT,30,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+        ENTRY(DEFAULT,0,ERROR_ACTION)
+
+};
+
+/*
+** State 29 - process binary point exponent sign
+*/
+static short int S29[] = {
+	ENTRY(DIGIT,30,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(ODIGIT,30,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+        ENTRY(DEFAULT,0,ERROR_ACTION)
+};
+
+/*
+** State 30 - recognize binary point exponent
+*/
+static short int S30[] = {
+        ENTRY(DIGIT,30,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+        ENTRY(ODIGIT,30,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+        ENTRY(DEFAULT,31,BUILD_ACTION)
+};
+
+/*
+** State 31 - build nnEnnBnn binary point number
+*/
+static short int S31[] = {
+        ENTRY(DEFAULT,0,NULL_ACTION)
+};
+
+/*
+** State 32 - recognize nnBnnEnn format
+*/
+static short int S32[] = {
+	ENTRY(PLUS,33,EAT_ACTION|BUILD_ACTION),
+	ENTRY(MINUS,33,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(EE,32,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(DIGIT,34,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(ODIGIT,34,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+        ENTRY(DEFAULT,0,ERROR_ACTION)
+};
+
+/*
+** State 33 - process binary point exponent sign
+*/
+static short int S33[] = {
+	ENTRY(DIGIT,34,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+	ENTRY(ODIGIT,34,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+        ENTRY(DEFAULT,0,ERROR_ACTION)
+};
+
+/*
+** State 34 - recognize binary point exponent
+*/
+static short int S34[] = {
+        ENTRY(DIGIT,34,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+        ENTRY(ODIGIT,34,EAT_ACTION|MOVE_ACTION|BUILD_ACTION),
+        ENTRY(DEFAULT,35,BUILD_ACTION)
+};
+
+/*
+** State 35 - build nnBnnEnn binary point number
+*/
+static short int S35[] = {
+        ENTRY(DEFAULT,0,NULL_ACTION)
+};
+
+
 static short int *scantable[] = {
          S1,  S2,  S3,  S4,  S5,  S6,  S7,  S8,  S9, S10,
 	S11, S12, S13, S14, S15, S16, S17, S18, S19, S20,
-	S21, S22, S23, S24, S25, S26, S27
+	S21, S22, S23, S24, S25, S26, S27, S28, S29, S30,
+	S31, S32, S33, S34, S35
 };
 
